#ifndef INCLUDED_SG_CONTROLLER_HXX
#define INCLUDED_SG_CONTROLLER_HXX

// $Id: sg_Controller.hxx,v 1.2 1997/02/05 20:18:25 moyers Exp $

#ifndef INCLUDED_SG_TYPES_HXX
#include "scenegraph/sg_Types.hxx"
#endif

class sg_Controller
{
   public:

      sg_Controller();

      virtual
      void sync();

      virtual
      void finish();
         // Deprecated.

      virtual
      void activate();
         // Called when the controller becomes the activate, or current controller.

      virtual
      void deactivate();
         // Called just before the controller is about to become inactive.

      enum KeyStatus {noKeysPressed = 0,
                      controlKey = 1,
                      shiftKey   = 2,
                      altKey     = 4,
                      button1    = 8,
                      button2    = 16,
                      button3    = 32};

      // Mouse events.
      virtual
      void mouseMove(long x, long y, long status = noKeysPressed);

      virtual
      void button1Down(long x, long y, long status = noKeysPressed);

      virtual
      void button1Up(long x, long y, long status = noKeysPressed);

      virtual
      void button2Down(long x, long y, long status = noKeysPressed);

      virtual
      void button2Up(long x, long y, long status = noKeysPressed);

      virtual
      void button3Down(long x, long y, long status = noKeysPressed);

      virtual
      void button3Up(long x, long y, long status = noKeysPressed);

      // Keyboard events other than modifier keys.
      virtual
      void keyPressed(char charRep, long status = noKeysPressed);

      virtual
      void keyReleased(char charRep, long status = noKeysPressed);
};


class sg_CameraNode;
class sg_SceneNode;

class sg_ControlLineController : public sg_Controller
{
   // Object movement similiar to control line model airplanes.
   // The object's local +Y axis is made "point" at an origin. The object then
   // moves around the (fixed) origin.

   public:
	   double getCurrentAltitude();
	   double getCurrentRadius();
	   void setCamera(sg_CameraNode const* camera);
      sg_ControlLineController(sg_CameraNode const* camera);
         // We need a camera so that we can get viewport parameters and other
         // camera specific stuff.

      void setNode(sg_SceneNode* node);
         // The node to be moved by the controller.
         // Default: 0 (NULL)

      void setOrigin(double x, double y, double z);
         // The point the camera points towards, and orbits.
         // Default: (0,0,0)

      void setMinRadius(double minRadius);
         // Default: 0

      void setMaxRadius(double maxRadius);
         // Default: 100

      void setCurrentRadius(double curRadius);
         // Default: 100

      void setCurrentAzimuth(double azimuth);
         // Default: 180

      void setCurrentAltitude(double altitude);
         // Default: 0

      double getCurrentAzimuth() const;

      double getGeoAzimuth() const;
      void setGeoAzimuth(double geoAzim);
         // +y axis = zero degrees

      void sync();
      void mouseMove(long x, long y, long status = noKeysPressed);
      void button1Up(long x, long y, long status = noKeysPressed);
      void button2Up(long x, long y, long status = noKeysPressed);

   private:

      sg_CameraNode const* _camera;
      sg_SceneNode* _node;
      sg_Point _origin;
      double _minRadius;
      double _maxRadius;
      double _curRadius;
      double _curAzimuth;
      double _curAltitude;
      bool _first_time;
      long _old_x, _old_y;

      void positionNode();
};

class sg_LightNode;
class sg_CompositeNode;
class sg_AltAzIndicator;
class sg_ShapeNode;

class sg_AltAzLightController : public sg_Controller
{
   public:


      sg_AltAzLightController(sg_CameraNode const* camera);
         // We need a camera so that we can get viewport parameters and other
         // camera specific stuff.

      void setLight(sg_LightNode* lightNode);
         // The light to be moved by the controller.
         // Default: 0 (NULL)

	   void setCamera(sg_CameraNode const* camera);

      void setOrigin(double x, double y, double z);
         // The point the camera points towards, and orbits.
         // Default: (0,0,0)

      void setRadius(double radius);
         // Default: 100

      double getGeoAzimuth();
      void setGeoAzimuth(double geoAzim);
         // +y axis = zero degrees

      void setCurrentAzimuth(double azimuth);
         // Default: 180

      void setCurrentAltitude(double altitude);
         // Default: 0

      double getCurrentAzimuth() const;

	   double getCurrentAltitude() const;

	   double getCurrentRadius() const;

      void sync();
      void mouseMove(long x, long y, long status = noKeysPressed);
      void button1Up(long x, long y, long status = noKeysPressed);
      void activate();
      void deactivate();

   private:

      sg_CameraNode const* _camera;
      sg_LightNode* _light;
      sg_Point _origin;
      double _radius;
      double _curAzimuth;
      double _curAltitude;
      bool _first_time;
      long _old_x, _old_y;
      sg_CompositeNode* _lightCompParent;
      sg_CompositeNode* _ourComposite;
      sg_AltAzIndicator* _altAzInd;
      sg_ShapeNode* _altAzIndNode;


      void positionNode();
};

class sg_DEM;

class sg_WaterLevelController : public sg_Controller
{
   public:

      sg_WaterLevelController(sg_CameraNode const* camera,
                              sg_DEM* demShape);

      void sync();

      void activate();
      void deactivate();

      void mouseMove(long x, long y, long status = noKeysPressed);
      void button1Up(long x, long y, long status = noKeysPressed);

   private:

      sg_CameraNode const* _camera;
      sg_DEM* _demShape;
      bool _first_time;
      double _dElev;
      long _old_x, _old_y;
      bool _prevWaterState;

};

#endif  // INCLUDED_SG_CONTROLLER_HXX
