// $Id: sg_DEM.cxx,v 1.8 1997/02/01 01:05:49 moyers Exp $


#include <math.h>

#ifdef WIN32
#else
#include <iostream.h>
#endif

#ifndef INCLUDED_SG_DEM_HXX
#include "scenegraph/sg_DEM.hxx"
#endif

#ifndef INCLUDED_SG_DEMVERTEXSHADER_HXX
#include "scenegraph/sg_DEMVertexShader.hxx"
#endif

#ifndef INCLUDED_HF_DEM_HXX
#include "heightfield/hf_DEM.hxx"
#endif

static
void
calcGouraudNormal(sg_Point const& p,
                  sg_Point const& p1,
                  sg_Point const& p2,
                  sg_Point const& p3,
                  sg_Point const& p4,
                  sg_Point const& p5,
                  sg_Point const& p6,
                  sg_Vector& normal)
{
   sg_Vector v1(p1.x - p.x, p1.y - p.y, p1.z - p.z);
   sg_Vector v2(p2.x - p.x, p2.y - p.y, p2.z - p.z);
   sg_Vector v3(p3.x - p.x, p3.y - p.y, p3.z - p.z);
   sg_Vector v4(p4.x - p.x, p4.y - p.y, p4.z - p.z);
   sg_Vector v5(p5.x - p.x, p5.y - p.y, p5.z - p.z);
   sg_Vector v6(p6.x - p.x, p6.y - p.y, p6.z - p.z);

   sg_Vector n1; n1 = v1.cross(v2); n1.normalize();
   sg_Vector n2; n2 = v2.cross(v3); n2.normalize();
   sg_Vector n3; n3 = v3.cross(v4); n3.normalize();
   sg_Vector n4; n4 = v4.cross(v5); n4.normalize();
   sg_Vector n5; n5 = v5.cross(v6); n5.normalize();
   sg_Vector n6; n6 = v6.cross(v1); n6.normalize();

   normal = n1 + n2 + n3 + n4 + n5 + n6;
   normal.normalize();
}                 

sg_DEM::sg_DEM(hf_DEM const* heightfield)
      : _heightfield(heightfield),
        _shader(0),
        _res(1.0),
        _gouraudFlag(false),
        _drawBoundingBox(false),
        _boundBoxColor(sg_RGBA(1.0, 1.0, 1.0)),
        _drawDEM(true),
        _waterColor(sg_RGBA(0.0, 0.3, 1.0, 0.5)),
        _drawWater(false),
        _transparentWater(false),
        _transAlpha(0.5)
{
   _numColumns = _heightfield->getWidth();
   _numRows = _heightfield->getHeight();
   _curRenderSteps = _numRows;
   _maxRenderSteps = _numRows;
   _missingDataValue = _heightfield->getMissingData();

   double dx = _heightfield->getXRes();
   double dy = _heightfield->getYRes();

   // World units are in meters. If the DEM is in arc-seconds, use an
   // approximation to the number of meters per arc second.

   if (_heightfield->getGroundRefSysUnits() == 3)
      {
         double metersPerArcSecond = 30.922;
         dx = dx * metersPerArcSecond;
         dy = dy * metersPerArcSecond;
      }
   if (_heightfield->getWorldUnits() == hf_DEM::FEET)       // check if feet
      {
         _convFactor = 0.3048006;
      }
   else                                                     // must be in meters
      {
         _convFactor = 1.0;
      }

   _width = _numColumns * dx;
   _height = _numRows * dy;
   _x_min = -1.0 * (_width / 2.0);
   _x_max = _x_min + _width;
   _y_min = -1.0 * (_height / 2.0);
   _y_max = _y_min + _height;

   _maxElev = _heightfield->getMaxElev() * _convFactor;
   _minElev = _heightfield->getMinElev() * _convFactor;
   _waterLevel = _minElev;

   setVerticalExageration(1.0);
}

hf_DEM const*
sg_DEM::getHeightField() const
{
   return _heightfield;
}


void 
sg_DEM::getDEMPoint(long virtRow, long virtCol, sg_Point& p)
{
   double r = (virtRow / (_virtRows * 1.0)) * _numRows;
   double c = (virtCol / (_virtColumns * 1.0)) * _numColumns;
   double elev = _heightfield->getElev(floor((c + 0.5)), floor((r + 0.5)));
   
   p.x = _x_min + (virtCol * _dx);
   p.y = _y_min + ((_virtRows - virtRow) * _dy);

   if (elev == _missingDataValue)
      p.z = elev;
   else
      p.z = _z_min + (((elev * _convFactor) - _minElev) * _vertExag);
}

void
sg_DEM::setupShaderInfo(long row, long column,  // virtual row and column of 'point'.
                        sg_Point const& point,
                        sg_Point const& q,
                        sg_Point const& r,
                        sg_DEMVertexShaderInfo& shaderInfo)
{
   // The 'q' and the 'r' explained.
   // 
   // 'point' is the point for which we're setting up the shader info structure.
   // q and r are also points in the DEM and have the following relationship
   // to point:
   //
   //   q 
   //
   //         point
   //
   //   r
   //
   // If we're not doing Gouraud normals (averaging the normals of the surrounding
   // triangles), then the normal at 'point' is:
   // vector(point,q) x vector(point,r).


   shaderInfo.q = q;
   shaderInfo.r = r;
   if (shaderInfo.q.z == _missingDataValue)
      shaderInfo.q.z = point.z;
   if (shaderInfo.r.z == _missingDataValue)
      shaderInfo.r.z = point.z;

   shaderInfo.vertex = point;

   if (_gouraudFlag)
   {
      sg_Point q1,q2,q3,q4,q5,q6;
      getDEMPoint(row - 1, column, q1);
      getDEMPoint(row, column - 1, q2);
      getDEMPoint(row + 1, column - 1, q3);
      getDEMPoint(row + 1, column, q4);
      getDEMPoint(row, column + 1, q5);
      getDEMPoint(row - 1, column + 1, q6);
      if (q1.z == _missingDataValue) q1.z = point.z;
      if (q2.z == _missingDataValue) q2.z = point.z;
      if (q3.z == _missingDataValue) q3.z = point.z;
      if (q4.z == _missingDataValue) q4.z = point.z;
      if (q5.z == _missingDataValue) q5.z = point.z;
      if (q6.z == _missingDataValue) q6.z = point.z;
      calcGouraudNormal(point, q1, q2, q3, q4, q5, q6, shaderInfo.normal);
   }
   else
   {
      // normal  = vec(p,shaderInfo.q) X vec(p,shaderInfo.r)
   
      sg_Vector v1(shaderInfo.q.x - point.x,
                   shaderInfo.q.y - point.y,
                   shaderInfo.q.z - point.z);

      sg_Vector v2(shaderInfo.r.x - point.x,
                   shaderInfo.r.y - point.y,
                   shaderInfo.r.z - point.z);

      shaderInfo.normal = v1.cross(v2);
      shaderInfo.normal.normalize();
   }
}


void
sg_DEM::getBoundingBox(double& min_x, double& min_y, double& min_z,
                       double& max_x, double& max_y, double& max_z) const
{
   min_x = _x_min;
   max_x = _x_max;
   min_y = _y_min;
   max_y = _y_max;
   min_z = _z_min;
   max_z = _z_max;
}

void
sg_DEM::getBoundingBox(sg_MinBoundBox& mbb) const
{
   mbb.min_x = _x_min;
   mbb.max_x = _x_max;
   mbb.min_y = _y_min;
   mbb.max_y = _y_max;
   mbb.min_z = _z_min;
   mbb.max_z = _z_max;
}


long
sg_DEM::render(long previousStep)
{
   if (_heightfield == 0)
      return 0;

   if (previousStep == 0)  // Setup for new rendering cycle.
      {
         if (_drawBoundingBox)
            drawBoundingBox();

         //if (_drawWater == true)
         //   drawWater();

         if (_drawDEM == false)
            return 0;

         _chunkSize = floor((_maxRenderSteps * 1.0) / _curRenderSteps);
         _curRow = 0;

         if (_shader != 0)
            {
               _shader->setExtent(_x_min, _x_max,
                                  _y_min, _y_max,
                                  _z_min, _z_max);
               _shader->init();
            }

         // Grab the ModelView and Projection matrices.
         glGetDoublev(GL_MODELVIEW_MATRIX,_modelViewMat);
         glGetDoublev(GL_PROJECTION_MATRIX,_projMat);
      }

   // The following code generates one or more triangle strips. It's 
   // complicated because it does not emit points for "void areas"
   // (areas where the elevation value for a vertex is equal to
   // the missing data value).
   //
   // It takes two DEM rows to make a triangle strip:
   //    E(n,0)   E(n,1)   E(n,2)   E(n,3)  ...
   //    E(n+1,0) E(n+1,1) E(n+1,2) E(n+1,3) ...
   // Where E(i,j) is the elevation value at row i, column j.
   // 
   // For the following discussion, lets label these points
   // as follows:
   //    1        3        5        7    ...
   //    2        4        6        8    ...
   // If all points were "good" (non-void), then the points
   // would be added to the OpenGL triangle strip in the order
   // they are labeled. Easy. :)  Things aren't quite so neat 
   // when one or more of those points are considered "bad" 
   // and should be skipped.
   //
   // It would be easy if we'd simply end the current triangle strip 
   // and start a new one with the next good point we encounter. 
   // Unfortunately, there is a lot of overhead involved in starting
   // a strip, so we'd like to keep the number of strips to a minumum.
   // This means we need some way to "skip" over triangles in a strip.
   // To do this, we can take advantage of the fact that OpenGL will
   // not draw zero-area triangles(1). A zero area triangle is one
   // in which at least two of the vertices are the same. 
   //
   // The trick then, is to know when to add the same point twice
   // to start a zero area triangle. 

   // (1) Paul Gray let me in on this little secret, which he got
   //     from Brian Paul (the author of Mesa). Thanks!




   sg_DEMVertexShaderInfo aShaderInfo, bShaderInfo;
   sg_Point a,b,lastPoint;
   bool stripStarted;
   bool skipColumn;

   long endRow = _curRow + _chunkSize;
   if (endRow > _virtRows - 1)
      endRow = _virtRows - 1;
   for (long row = _curRow; row < endRow; row++)
   {
      stripStarted = false;
      skipColumn = false;

      glBegin(GL_TRIANGLE_STRIP);
      for (long column = 0; column < _virtColumns; column++)
      {
         getDEMPoint(row, column, a);
         getDEMPoint(row + 1, column, b);

         sg_Point r;
         getDEMPoint(row + 1, column - 1, r);
         if (a.z != _missingDataValue)
         {
            sg_Point q;
            getDEMPoint(row, column - 1, q);
            setupShaderInfo(row, column, a, q, r, aShaderInfo);
         }

         if (b.z != _missingDataValue)
         {
            setupShaderInfo(row + 1, column, b, a, r, bShaderInfo);
         }

         if (a.z != _missingDataValue)
         {
            if (skipColumn)
            {
               glVertex3d(a.x, a.y, a.z);
               skipColumn = false;
            }

            if (_shader != 0)
            {
               _shader->shade(aShaderInfo);
               a = aShaderInfo.vertex;
            }
            glVertex3d(a.x, a.y, a.z);

            if (b.z != _missingDataValue)
            {
               if (_shader != 0)
               {
                  _shader->shade(bShaderInfo);
                  b = bShaderInfo.vertex;
               }
               glVertex3d(b.x, b.y, b.z);
               lastPoint = b;
            }
            else
            {
               glVertex3d(a.x, a.y, a.z);
               lastPoint = a;
            }
            stripStarted = true;
         }
         else 
            if (b.z != _missingDataValue)
            {
               if (skipColumn)
               {
                  glVertex3d(b.x, b.y, b.z);
                  skipColumn = false;
               }
               if (_shader != 0)
               {
                  _shader->shade(bShaderInfo);
                  b = bShaderInfo.vertex;
               }
               glVertex3d(b.x, b.y, b.z);
               glVertex3d(b.x, b.y, b.z);
               lastPoint = b;
               stripStarted = true;
            }
            else
               if (!skipColumn && stripStarted)
               {
                  glVertex3d(lastPoint.x, lastPoint.y, lastPoint.z);
                  skipColumn = true;
               }

      }
      glEnd();
   }

  _curRow = endRow;

   if ( (_curRenderSteps - _curRow -1) == 0)
   {
	   _shader->cleanup();
      if (_drawWater == true)
         drawWater();
   }

   return _curRenderSteps - _curRow -1;
}


void
sg_DEM::setShader(sg_DEMVertexShader* shader)
{
	if (_shader != NULL) _shader->cleanup();
	_shader = shader;
}

void
sg_DEM::setResolution(double res)
{
   _res = res;
   if (_res < 0.0)
      _res = 0.0;
   if (_res > 1.0)
      _res = 1.0;

   _virtColumns = floor(_numColumns * _res);
   if (_virtColumns < 2)
      _virtColumns = 2;   

   _virtRows = floor(_numRows * _res);
   if (_virtRows < 2)
      _virtRows = 2;

   _dx = (_x_max - _x_min) / _virtColumns;
   _dy = (_y_max - _y_min) / _virtRows;

   _maxRenderSteps = _virtRows;
   _curRenderSteps = _virtRows;
}

void
sg_DEM::setVerticalExageration(double ve)
{
   _vertExag = ve;

   double dz = (_maxElev - _minElev) * _vertExag;
   _z_min = -1.0 * (dz / 2.0);
   _z_max = _z_min + dz;
}

double
sg_DEM::getVerticalExageration() const
{
   return _vertExag;
}

double
sg_DEM::getMinElev() const
{
   return _minElev;
}

double
sg_DEM::getMaxElev() const
{
   return _maxElev;
}

void
sg_DEM::setGouraud(bool flag)
{
   _gouraudFlag = flag;
}

bool
sg_DEM::isGouraud() const
{
   return _gouraudFlag;
}

void
sg_DEM::getModelView(GLdouble* modelView)
{
   for (long i = 0; i < 16; i++)
      modelView[i] = _modelViewMat[i];
}

void
sg_DEM::getProjection(GLdouble* projection)
{
   for (long i = 0; i < 16; i++)
      projection[i] = _projMat[i];
}

void
sg_DEM::setDrawBoundingBox(bool flag)
{
   _drawBoundingBox = flag;
}

void
sg_DEM::setBoundingBoxColor(sg_RGBA const& bbColor)
{
   _boundBoxColor = bbColor;
}

void
sg_DEM::drawBoundingBox()
{
   glDisable(GL_LIGHTING);
   glColor3d(_boundBoxColor.red, _boundBoxColor.green, _boundBoxColor.blue);

   glEnable(GL_LINE_STIPPLE);
   glLineStipple(1, 0x00FF);
   glBegin(GL_LINE_STRIP);
      glVertex3d(_x_max, _y_max, _z_max);
      glVertex3d(_x_min, _y_max, _z_max);
      glVertex3d(_x_min, _y_min, _z_max);
      glVertex3d(_x_max, _y_min, _z_max);
      glVertex3d(_x_max, _y_max, _z_max);
      glVertex3d(_x_max, _y_max, _z_min);
      glVertex3d(_x_min, _y_max, _z_min);
      glVertex3d(_x_min, _y_min, _z_min);
      glVertex3d(_x_max, _y_min, _z_min);
      glVertex3d(_x_max, _y_max, _z_min);
   glEnd();

   glBegin(GL_LINES);
      glVertex3d(_x_min, _y_max, _z_max);
      glVertex3d(_x_min, _y_max, _z_min);

      glVertex3d(_x_min, _y_min, _z_max);
      glVertex3d(_x_min, _y_min, _z_min);

      glVertex3d(_x_max, _y_min, _z_max);
      glVertex3d(_x_max, _y_min, _z_min);
   glEnd();

   glDisable(GL_LINE_STIPPLE);
   glEnable(GL_LIGHTING);
}

void
sg_DEM::setDrawDEM(bool flag)
{
   _drawDEM = flag;
}

void
sg_DEM::setWaterColor(sg_RGBA const& waterColor)
{
   _waterColor = waterColor;
}

sg_RGBA const&
sg_DEM::getWaterColor() const
{
   return _waterColor;
}

void
sg_DEM::drawWater()
{
   double wz = ((_waterLevel - _minElev) * _vertExag) + _z_min;

   double xm = _x_min * 1.025;
   double ym = _y_min * 1.025;
   double zm = _z_min * 1.025;

   double xM = _x_max * 1.025;
   double yM = _y_max * 1.025;

   if (_transparentWater)
   {
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   }

   glColor4f(_waterColor.red, _waterColor.green, _waterColor.blue, 
             _transAlpha);

   glFrontFace(GL_CCW);
   glEnable(GL_CULL_FACE);
   glBegin(GL_QUADS);
      glNormal3f(0.0, 0.0, 1.0);
      glVertex3d(xM, yM, wz);
      glVertex3d(xm, yM, wz);
      glVertex3d(xm, ym, wz);
      glVertex3d(xM, ym, wz);

      glNormal3f(0.0, 0.0, -1.0);
      glVertex3d(xM, yM, zm);
      glVertex3d(xM, ym, zm);
      glVertex3d(xm, ym, zm);
      glVertex3d(xm, yM, zm);
   glEnd();

   glBegin(GL_QUAD_STRIP);
      glNormal3f(0.0, 1.0, 0.0);
      glVertex3d(xM, yM, wz);
      glVertex3d(xM, yM, zm);
      glVertex3d(xm, yM, wz);
      glVertex3d(xm, yM, zm);

      glNormal3f(-1.0, 0.0, 0.0);
      glVertex3d(xm, ym, wz);
      glVertex3d(xm, ym, zm);
      
      glNormal3f(0.0, -1.0, 0.0);
      glVertex3d(xM, ym, wz);
      glVertex3d(xM, ym, zm);
      
      glNormal3f(1.0, 0.0, 0.0);
      glVertex3d(xM, yM, wz);
      glVertex3d(xM, yM, zm); 
   glEnd();

   glDisable(GL_CULL_FACE);

   if (_transparentWater)
      glDisable(GL_BLEND);
}

void
sg_DEM::setDrawWater(bool flag)
{
   _drawWater = flag;
}

bool
sg_DEM::getDrawWater() const
{
   return _drawWater;
}

void
sg_DEM::setWaterLevel(double waterZ)
{
   if (waterZ > _maxElev)
      waterZ = _maxElev;
   else
      if (waterZ < _minElev)
         waterZ = _minElev;

   _waterLevel = waterZ;
}

double
sg_DEM::getWaterLevel() const
{
   return _waterLevel;
}

void
sg_DEM::setTransparentWater(bool flag)
{
   _transparentWater = flag;
}

bool
sg_DEM::getTransparentWater() const
{
   return _transparentWater;
}

void
sg_DEM::setTransAlpha(double transAlpha)
{
   if (transAlpha < 0.0)
      transAlpha = 0.0;
   else
      if (transAlpha > 1.0)
         transAlpha = 1.0;

   _transAlpha = transAlpha;
}

double
sg_DEM::getTransAlpha() const
{
   return _transAlpha;  
}