#ifndef INCLUDED_SG_DEMVERTEXSHADER_HXX
#define INCLUDED_SG_DEMVERTEXSHADER_HXX

// $Id: sg_DEMVertexShader.hxx,v 1.5 1997/01/29 19:25:13 pjgray Exp $

#ifdef WIN32
#include <afxwin.h>
#include <afxext.h>
#include <afxtempl.h>
#include <vector>
using namespace std;
#else
#include <vector.h>
#endif

#ifndef INCLUDED_SG_TYPES_HXX
#include "scenegraph/sg_Types.hxx"
#endif

#ifndef INCLUDED_SG_VECTOR_HXX
#include "scenegraph/sg_Vector.hxx"
#endif

class hf_DEM;

struct sg_DEMVertexShaderInfo
{
   sg_Point vertex;
   sg_Point q;
   sg_Point r;
   sg_Vector normal;   // vec(vertex,q) x vec(vertex,r);
};

class sg_DEMVertexShader
{
   public:

      sg_DEMVertexShader(hf_DEM const* heightField);

      void setExtent(double x_min, double x_max,
                     double y_min, double y_max,
                     double z_min, double z_max);
         // Sets the bounding box of the DEM.

      virtual
      void init();

	  virtual
	  void cleanup();

      virtual
      void shade(sg_DEMVertexShaderInfo& info);
         // Calculates a current vertex color and  material color.

   protected:
      hf_DEM const* _heightfield;
      double _x_min, _x_max;
      double _y_min, _y_max;
      double _z_min, _z_max;
};

class sg_DaylightShader : public sg_DEMVertexShader
{
   // Diffuse reflection based on the dot product of the surface normal and 
   // direction to light. Basically, just calculates a surface normal.

   public:

      sg_DaylightShader(hf_DEM const* heightField);

      void shade(sg_DEMVertexShaderInfo& info);

      void setColor(sg_RGBA const& color);

      sg_RGBA const& getColor() const;

   private:

      sg_RGBA _demColor;  // Default (0.8, 0.8, 0.8, 1.0)
};


class sg_HSVShader : public sg_DEMVertexShader
{
   // HSV ramp between min and max elevation.

   public:

      sg_HSVShader(hf_DEM const* heightField);
      void init();
      void shade(sg_DEMVertexShaderInfo& info);

      void setLowColorStart(double deg);
         // 0.0 <= deg <= 360.0.  Controls where the color mapped to the
         // lowest elevation starts in HSV space.
         // Default: 0.0

      void setSaturation(double sat);
         // 0.0 <= sat <= 1.0
         // Default: 1.0

      void setValue(double val);
         // 0.0 <= val <= 1.0
         // Default: 1.0

      void setRange(double range);
         // 1.0 < range;
         // Default: 360.0

      double getLowColorStart() const;
      double getSaturation() const;
      double getValue() const;
      double getRange() const;

   private:
      double _lowColorStart;
      double _saturation;
      double _value;
      double _range;
};

class sg_RangeShader : public sg_DEMVertexShader
{
   public:

      sg_RangeShader(hf_DEM const* heightField);
      void shade(sg_DEMVertexShaderInfo& info);

      void setColors(vector<sg_RGBA> const& colorlist);
         // List of colors, from low altitude to high altitude.

   private:

      vector<sg_RGBA> _colors;
};

class sg_SlopeShader : public sg_DEMVertexShader
{
   public:

      sg_SlopeShader(hf_DEM const* heightField);
      void init();
      void shade(sg_DEMVertexShaderInfo& info);

      void setFlatColor(sg_RGBA const& flatColor);
      void setSteepColor(sg_RGBA const& steepColor);

      sg_RGBA const& getFlatColor() const;
      sg_RGBA const& getSteepColor() const;

   private:
      sg_RGBA _flatColor;
      sg_RGBA _steepColor;
      double _red_diff;
      double _green_diff;
      double _blue_diff;
};

class sg_HeadingShader : public sg_DEMVertexShader
{

   public:

      sg_HeadingShader(hf_DEM const* heightField);

      void shade(sg_DEMVertexShaderInfo& info);
};


class sg_ContourShader : public sg_DEMVertexShader
{
   // A displacement shader. It snaps the z value of the point to
   // some user specified tolerance.

   public:

      sg_ContourShader(hf_DEM const* heightField);

      void shade(sg_DEMVertexShaderInfo& info);

      void setColor1(sg_RGBA const& color1);
      void setColor2(sg_RGBA const& color2);
      void setEdgeColor(sg_RGBA const& edgeColor);

      sg_RGBA const& getColor1();
      sg_RGBA const& getColor2();
      sg_RGBA const& getEdgeColor();

      void setContourInterval(double interval);
         // Interval specified in world coordinates.
         // Default: 10
      double getContourInterval() const;

   private:
      double _interval;
	   sg_RGBA _color1;     // Default: sg_Color_Blue
	   sg_RGBA _color2;     // Default: sg_Color_Yellow
      sg_RGBA _edgeColor;  // Default: sg_Color_Red
};

class sg_GradientShader : public sg_DEMVertexShader
{

   // A gradient shader. Shades between a "low" color and a "high" color.

   public:

      sg_GradientShader(hf_DEM const* heightField);

      void shade(sg_DEMVertexShaderInfo& info);

      void setLowColor(sg_RGBA const& lowColor);
      void setHighColor(sg_RGBA const& highColor);

      sg_RGBA const& getLowColor() const;
      sg_RGBA const& getHighColor() const;

   private:

      sg_RGBA _lowColor;
      sg_RGBA _highColor;

};

class sg_LakeShader : public sg_DEMVertexShader
{

   // A hydrographic shader. Shades flat areas a single color. Shades everything
   // else a different color.

   public:

      sg_LakeShader(hf_DEM const* heightField);

      void shade(sg_DEMVertexShaderInfo& info);

      void setLakeColor(sg_RGBA const& lakeColor);
      void setLandColor(sg_RGBA const& landColor);

      sg_RGBA const& getLakeColor() const;
      sg_RGBA const& getLandColor() const;

   private:

      sg_RGBA _lakeColor;
      sg_RGBA _landColor;

};

class sg_TextureShader : public sg_DEMVertexShader
{
   // My attempt at a texture shader

   public:

      sg_TextureShader(hf_DEM const* heightField);
      void shade(sg_DEMVertexShaderInfo& info);
	  void init();
	  void cleanup();
	  void setImageWidth(long width);
	  void setImageHeight(long height);
	  void setDRGWidth(long drgWidth);
	  void setDRGHeight(long drgHeight);
	  void setPixels(GLuint* rgb);

   private:
      double _imageWidth;
      double _imageHeight;
	  double _drgHeight;
	  double _drgWidth;
      GLuint *_pixels;
};



#endif // INCLUDED_SG_DEMVERTEXSHADER_HXX
