
// $Id: sg_MatrixUtils.cxx,v 1.3 1997/01/29 19:25:14 pjgray Exp $

#include <math.h>

#ifdef WIN32
#define M_PI 3.14159265358979323846
#endif

#ifndef INCLUDED_SG_MATRIXUTILS_HXX
#include "scenegraph/sg_MatrixUtils.hxx"
#endif

#ifndef INCLUDED_SG_MATRIX_HXX
#include "scenegraph/sg_Matrix.hxx"
#endif

#ifndef INCLUDED_SG_VECTOR_HXX
#include "scenegraph/sg_Vector.hxx"
#endif

#ifndef INCLUDED_SG_TYPES_HXX
#include "scenegraph/sg_Types.hxx"
#endif

void
sg_MatrixUtils::rotateMatrix(sg_Matrix& matrix,
                             double angle, double x, double y, double z)
{
   sg_Matrix rotmat;
   sg_MatrixUtils::genRotateMatrix(angle,x,y,z,rotmat);
   matrix = matrix * rotmat;
}


void
sg_MatrixUtils::translateMatrix(sg_Matrix& matrix,
                                double dx, double dy, double dz)
{
   double* m = matrix._matrix;

   m[12] = m[0] * dx + m[4] * dy + m[8]  * dz + m[12];
   m[13] = m[1] * dx + m[5] * dy + m[9]  * dz + m[13];
   m[14] = m[2] * dx + m[6] * dy + m[10] * dz + m[14];
   m[15] = m[3] * dx + m[7] * dy + m[11] * dz + m[15];
}

void
sg_MatrixUtils::scaleMatrix(sg_Matrix& matrix,
                            double sx, double sy, double sz)
{
   // From a function in Mesa's 'matrix.c'
   double* m = matrix._matrix;
   m[0] *= sx;   m[4] *= sy;   m[8]  *= sz;
   m[1] *= sx;   m[5] *= sy;   m[9]  *= sz;
   m[2] *= sx;   m[6] *= sy;   m[10] *= sz;
   m[3] *= sx;   m[7] *= sy;   m[11] *= sz;
}


void
sg_MatrixUtils::transformPoint(sg_Matrix const& matrix,
                                  sg_Point const& p,
                                  sg_Point& q)
{
   double const* m = matrix._matrix;
   q.x = p.x * m[0] + p.y * m[4] + p.z * m[8] + m[12];
   q.y = p.x * m[1] + p.y * m[5] + p.z * m[9] + m[13];
   q.z = p.x * m[2] + p.y * m[6] + p.z * m[10] + m[14];
}

void
sg_MatrixUtils::genRotateMatrix(double angle, double x, double y, double z,
                                sg_Matrix& matrix)
{
   // From a function in Mesa's matrix.c

   double mag, s, c;
   double xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;

   s = sin( angle * (M_PI / 180.0) );
   c = cos( angle * (M_PI / 180.0) );

   mag = sqrt( x*x + y*y + z*z );

   if (mag == 0.0)
     return;

   x /= mag;
   y /= mag;
   z /= mag;


   /*
    *     Arbitrary axis rotation matrix.
    *
    *  This is composed of 5 matrices, Rz, Ry, T, Ry', Rz', multiplied
    *  like so:  Rz * Ry * T * Ry' * Rz'.  T is the final rotation
    *  (which is about the X-axis), and the two composite transforms
    *  Ry' * Rz' and Rz * Ry are (respectively) the rotations necessary
    *  from the arbitrary axis to the X-axis then back.  They are
    *  all elementary rotations.
    *
    *  Rz' is a rotation about the Z-axis, to bring the axis vector
    *  into the x-z plane.  Then Ry' is applied, rotating about the
    *  Y-axis to bring the axis vector parallel with the X-axis.  The
    *  rotation about the X-axis is then performed.  Ry and Rz are
    *  simply the respective inverse transforms to bring the arbitrary
    *  axis back to it's original orientation.  The first transforms
    *  Rz' and Ry' are considered inverses, since the data from the
    *  arbitrary axis gives you info on how to get to it, not how
    *  to get away from it, and an inverse must be applied.
    *
    *  The basic calculation used is to recognize that the arbitrary
    *  axis vector (x, y, z), since it is of unit length, actually
    *  represents the sines and cosines of the angles to rotate the
    *  X-axis to the same orientation, with theta being the angle about
    *  Z and phi the angle about Y (in the order described above)
    *  as follows:
    *
    *  cos ( theta ) = x / sqrt ( 1 - z^2 )
    *  sin ( theta ) = y / sqrt ( 1 - z^2 )
    *
    *  cos ( phi ) = sqrt ( 1 - z^2 )
    *  sin ( phi ) = z
    *
    *  Note that cos ( phi ) can further be inserted to the above
    *  formulas:
    *
    *  cos ( theta ) = x / cos ( phi )
    *  sin ( theta ) = y / sin ( phi )
    *
    *  ...etc.  Because of those relations and the standard trigonometric
    *  relations, it is pssible to reduce the transforms down to what
    *  is used below.  It may be that any primary axis chosen will give the
    *  same results (modulo a sign convention) using thie method.
    *
    *  Particularly nice is to notice that all divisions that might
    *  have caused trouble when parallel to certain planes or
    *  axis go away with care paid to reducing the expressions.
    *  After checking, it does perform correctly under all cases, since
    *  in all the cases of division where the denominator would have
    *  been zero, the numerator would have been zero as well, giving
    *  the expected result.
    */

   xx = x * x;
   yy = y * y;
   zz = z * z;
   xy = x * y;
   yz = y * z;
   zx = z * x;
   xs = x * s;
   ys = y * s;
   zs = z * s;
   one_c = 1.0F - c;

   matrix(0,0) = (one_c * xx) + c;
   matrix(0,1) = (one_c * xy) - zs;
   matrix(0,2) = (one_c * zx) + ys;
   matrix(0,3) = 0.0F;

   matrix(1,0) = (one_c * xy) + zs;
   matrix(1,1) = (one_c * yy) + c;
   matrix(1,2) = (one_c * yz) - xs;
   matrix(1,3) = 0.0F;

   matrix(2,0) = (one_c * zx) - ys;
   matrix(2,1) = (one_c * yz) + xs;
   matrix(2,2) = (one_c * zz) + c;
   matrix(2,3) = 0.0F;

   matrix(3,0) = 0.0F;
   matrix(3,1) = 0.0F;
   matrix(3,2) = 0.0F;
   matrix(3,3) = 1.0F;

}

void
sg_MatrixUtils::invertMatrix(sg_Matrix const& b,
                             sg_Matrix& a)
{
   // Comments from the version stolen from Mesa.
/*
 * Find the inverse of the 4 by 4 matrix b using gausian elimination
 * and return it in a.
 *
 * This function was contributed by Thomas Malik (malik@rhrk.uni-kl.de).
 * Thanks Thomas!
 */

  double val, val2;
  long   i, j, k, ind;
  sg_Matrix tmp;

  a.identity();
  tmp = b;

  for (i = 0; i != 4; i++) {
    val = tmp(i,i);   // Find pivot
    ind = i;
    for (j = i + 1; j != 4; j++) {
      if (fabs(tmp(j,i)) > fabs(val)) {
        ind = j;
        val = tmp(j,i);
      }
    }

    if (ind != i) {                     /* swap columns */
      for (j = 0; j != 4; j++) {
        val2 = a(i,j);
        a(i,j) = a(ind,j);
        a(ind,j) = val2;
        val2 = tmp(i,j);
        tmp(i,j) = tmp(ind,j);
        tmp(ind,j) = val2;
      }
    }

    if (val == 0.0F) {  
       /* The matrix is singular (has no inverse).  This isn't really
        * an error since singular matrices can be used for projecting
        * shadows, etc.  We let the inverse be the identity matrix.
        */
       a.identity();
       return;
    }

    for (j = 0; j != 4; j++) {
      tmp(i,j) /= val;
      a(i,j) /= val;
    }

    for (j = 0; j != 4; j++) {          /* eliminate column */
      if (j == i)
        continue;
      val = tmp(j,i);
      for (k = 0; k != 4; k++) {
        tmp(j,k) -= tmp(i,k) * val;
        a(j,k) -= a(i,k) * val;
      }
    }
  }
}

void
sg_MatrixUtils::calcCameraMatrix(sg_Vector& up,
                                   sg_Vector& look,
                                   sg_Point const& p,
                                   sg_Vector& right,
                                   sg_Matrix& am)
{
   // We need a vector that will move the origin to the specified point
   // and line the world up with the given vectors. So, calculate a matrix
   // that will move the point to the origin and then take the inverse.


   // Y vector
   sg_Vector y(look);
   y.normalize();

   // Z vector
   sg_Vector z(up);   

   // X vector = Y cross Z
   sg_Vector x;
   x = y.cross(z);

   // Recompute Z = X cross Y
   z = x.cross(y);

   x.normalize();
   z.normalize();

   sg_Matrix m;
   x.getComponents(m(0,0), m(0,1), m(0,2));
   y.getComponents(m(1,0), m(1,1), m(1,2));
   z.getComponents(m(2,0), m(2,1), m(2,2));

   sg_MatrixUtils::translateMatrix(m, -1 * p.x, -1 * p.y, -1 * p.z);

   // m takes the camera to the origin. We need the matrix that takes the
   // origin to the camera. So find the inverse.

   sg_MatrixUtils::invertMatrix(m,am);

}
