#ifndef INCLUDED_SG_SCENENODE_HXX
#define INCLUDED_SG_SCENENODE_HXX

// $Id: sg_SceneNode.hxx,v 1.7 1997/02/01 01:03:12 moyers Exp $

#ifdef WIN32
#include <afxwin.h>
#include <afxext.h>
#include <afxtempl.h>
#include <vector>
using namespace std;
#else
#include <vector.h>
#endif

#ifndef INCLUDED_SG_SCENENODEPTR_HXX
#include "scenegraph/sg_SceneNodePtr.hxx"
#endif

#ifndef INCLUDED_SG_SCENEITERATOR_HXX
#include "scenegraph/sg_SceneIterator.hxx"
#endif

#ifndef INCLUDED_SG_TRANSFORMATION_HXX
#include "scenegraph/sg_Transformation.hxx"
#endif

#ifndef INCLUDED_SG_APPEARANCE_HXX
#include "scenegraph/sg_Appearance.hxx"
#endif

#ifndef INCLUDED_SG_TYPES_HXX
#include "scenegraph/sg_Types.hxx"
#endif

#ifndef INCLUDED_SG_VECTOR_HXX
#include "scenegraph/sg_Vector.hxx"
#endif

class sg_CompositeNode;
class sg_SceneVisitor;
class sg_Appearance;
class sg_Shape;

class sg_SceneNode
{
   public:

      sg_SceneNode();
      virtual ~sg_SceneNode();

      virtual
      void render();
         // Causes the node to "render". In other words, do what ever it does to
         // contribute to the rendering of the scene.

      virtual
      void renderLoRes();
         // Default Behavior: render()

      // Accessors
      sg_Transformation& getTransformation();
      sg_Shape* getShape();
      sg_Shape const* getShape() const;
      sg_SceneNode const* getParent() const;


      bool isActive() const;
         // Returns the active status of this node. If a node is not active,
         // visitors should ignore it.

      virtual
      bool getBoundingBox(double& min_x, double& min_y, double& min_z,
                          double& max_x, double& max_y, double& max_z) const;
         // If this node has a bounding box, (min_x, min_y, min_z), 
         // (max_x, max_y, max_z) will be set to the extreme corners of the box
         // and the function returns true. If this node has no bounding box
         // then it just returns false.

      virtual
      bool getBoundingBox(sg_MinBoundBox& mbb) const;

      virtual
      sg_CompositeNode* getComposite();
         // Returns the this ptr if this node is an sg_CompositeNode. Returns
         // false otherwise.

      sg_Appearance::ShadingStyle getShadingStyle() const;

      // Modifiers
      void setActive(bool activeFlag);
         // If true, visitors will process this node. If false,
         // visitors should ignore this node and it's children.

      void identity();
         // Sets node's transformation to the Identity transformation.

      void rotate(double angle, double x, double y, double z);
         // Rotation about a line from the origin to the point (x,y,z)
         // by angle 'angle'.

      void translate(double dx, double dy, double dz);

      void scale(double sx, double sy, double sz);

      void setShape(sg_Shape* shape);
      void setParent(sg_SceneNode* parent);

      // Appearance
      void setColor(sg_RGBA const& color);

      void setShadingStyle(sg_Appearance::ShadingStyle style);

      void setLighting(bool flag);

      bool isLit();

      void setPolygonMode(sg_Appearance::PolygonMode mode);

      sg_Appearance::PolygonMode getPolygonMode();

      void setShininess(double shininess);
      double getShininess() const;

      void setSpecularColor(sg_RGBA const& specColor);
      sg_RGBA const& getSpecularColor() const;

      // Visitor 
      virtual
      void accept(sg_SceneVisitor& visitor);

   private:
      bool _active;
      sg_SceneNode* _parent;
      sg_Transformation _transformation;
      sg_Appearance _appearance;
      sg_Shape* _shape;
};


class sg_ShapeNode : public sg_SceneNode
{
   public:

      sg_ShapeNode(sg_Shape* myshape = 0);

      void render();
      void renderLoRes();
      void accept(sg_SceneVisitor& visitor);

      bool getBoundingBox(double& min_x, double& min_y, double& min_z,
                          double& max_x, double& max_y, double& max_z) const;

      bool getBoundingBox(sg_MinBoundBox& mbb) const;

};

class sg_CompositeNode : public sg_SceneNode
{
   public:
      sg_CompositeNode();

      sg_CompositeNode* getComposite();

      void accept(sg_SceneVisitor& visitor);

      void addNode(sg_SceneNode* node);
         // Adds node nodeptr as a child of this node.

      void removeNode(sg_SceneNode* node);
         // Remove nodeptr from this node's set of children.

      void replaceNode(sg_SceneNode* oldnode,
                       sg_SceneNode* newnode);
         // Replaces oldnodeptr with newnodeptr.

   private:
      
      friend sg_SceneIterator;

      typedef vector<sg_SceneNodePtr> NodeVector;
      NodeVector _children;
};

class sg_LightNode : public sg_SceneNode
{
   public:

      sg_LightNode(GLenum lightID = GL_LIGHT0);
         // This light node will control one of the eight OpenGL lights. Provide
         // the ID of the light to be controlled (GL_LIGHT0, GL_LIGHT1, etc).

	  ~sg_LightNode();
		// Disables Light

      // Accessors

      GLenum getLightID() const;
      bool isOn() const;


      // Modifiers

      void render();
      void accept(sg_SceneVisitor& visitor);

      void turnOn();
      void turnOff();

      // Icon for light
      void setIcon(sg_ShapeNode* icon);
         // Light does not own the icon node, and will not delete it.
      void showIcon(bool flag);
         // If flag is true, the shape representing the light is drawn.
      bool iconOn() const;
         // Whether or not the light symbol is being drawn.

      void setPosition(sg_Point const& pos);
      void setPosition(double x, double y, double z);

      void setDirection(sg_Vector const& dir);
      void setDirection(double i, double j, double k);

      void shineOn(sg_Point const& target);
         // Direction will be a vector from current position to target.

      void setAmbient(sg_RGBA const& rgba);
      void setAmbient(double r, double g, double b, double a = 1.0);

      void setDiffuse(sg_RGBA const& rgba);
      void setDiffuse(double r, double g, double b, double a = 1.0);

      void setSpecular(sg_RGBA const& rgba);
      void setSpecular(double r, double g, double b, double a = 1.0);

      void setConstantAtten(double factor);
      void setLinearAtten(double factor);
      void setQuadraticAtten(double factor);

   protected:

      sg_Point _pos;           // Default: (0,0,0)
      sg_Vector _dir;          // Default: <0,1,0>

      GLenum  _lightID;         // Default: GL_LIGHT0
      bool    _on;              // Default: true
      sg_ShapeNode* _icon;
      bool    _showIcon;        // Default: false
      sg_RGBA _ambient;         // Default: (0,0,0,1)
      sg_RGBA _diffuse;         // Default: (1,1,1,1)
      sg_RGBA _specular;        // Default: (1,1,1,1)
      double  _constantAtten;   // Default: 1.0
      double  _linearAtten;     // Default: 0.0
      double  _quadAtten;       // Default: 0.0
};

class sg_PointLight : public sg_LightNode
{
   public:
      
      sg_PointLight(GLenum lightID);

      void render();
};

class sg_SpotLight : public sg_LightNode
{
   public:

      sg_SpotLight(GLenum lightID);

      void render();

      void setExponent(double exponent);
      void setCutoffAngle(double cutoffAngle);
         // angle between center line of light and edge of cone.
         // 0.0 <= coneAngle <= 90.0

   private:

      double _exponent;        // Default: 0.0
      double _cutoffAngle;     // Default: 45.0
};

class sg_DistantLight : public sg_LightNode
{
   public:

      sg_DistantLight(GLenum lightID);

      void render();
};

class sg_ManipulatorNode : public sg_SceneNode
{
   public:

      void accept(sg_SceneVisitor& visitor);
};

class sg_CameraNode : public sg_SceneNode
{
   public:

      sg_CameraNode();
      
      void render();
      void accept(sg_SceneVisitor& visitor);

      void look(sg_Point const& eye,
                sg_Point const& point,
                sg_Vector const& up);
      void look(double eye_x, double eye_y, double eye_z,
                double px, double py, double pz,
                double up_i, double up_j, double up_k);
         // Repositions the camera to be at (eye_x, eye_y, eye_z) and looking
         // at point (px,py,pz). <up_i,up_j,up_k> is toward the top of the screen.

      void setFieldOfView(double degrees);
         // Default: 45 degrees

      void setAspect(double ratio);
         // Default: 1.0

      void setFrontClipPlane(double dist);
         // Default: 1.0

      void setBackClipPlane(double dist);
         // Default: 10.0

      void setViewPort(long ll_x, long ll_y,      // Lower left corner of viewport
                       long width, long height);  // (all values in pixels).
         // Default: (0,0,100,100)

      double getFieldOfView() const;
      double getAspect() const;

      void getViewPort(long& ll_x, long& ll_y,
                       long& width, long& height) const;

      void getUpVector(sg_Vector& up) const;
      void getLookVector(sg_Vector& look) const;
      void getRightVector(sg_Vector& right) const;

   private:

      sg_Point _eye;        // Where we are.
      sg_Point _point;      // What we're looking at.
      sg_Vector _up;        // Which way is up. (See, I do *too* know... :-)
      sg_Vector _look;      // Direction we're looking (in direction of _point).
      sg_Vector _right;     // Calculated as _look X _up

      double _fieldOfView;
      double _aspect;
      double _frontClip;
      double _backClip;

      long _vpX, _vpY;    // Lower left (x,y) of viewport.
      long _vpWidth, _vpHeight;   // Width and Height of viewport.
};

#endif  // INCLUDED_SG_SCENENODE_HXX


