// $Id: sg_Shape.hxx,v 1.7 1997/01/29 19:39:17 moyers Exp $
#ifndef INCLUDED_SG_SHAPE_HXX
#define INCLUDED_SG_SHAPE_HXX

#ifdef WIN32
#include <afxwin.h>
#include <afxext.h>
#include <afxtempl.h>
#include <string>
using namespace std;
#else
#include <string>
#endif

#include <gl/gl.h>
#include <gl/glu.h>

#ifndef INCLUDED_SG_TYPES_HXX
#include "scenegraph/sg_Types.hxx"
#endif

class sg_Vector;

class sg_Shape
{
   // Issues commands that actually cause a shape to be drawn by OpenGL.
   public:
      sg_Shape();
      virtual ~sg_Shape();

      // Incremental rendering support
	   long getMaxRenderSteps();
	      // Returns the maximum number of steps that can be 
	      // taken to render the Shape.
	      // Default: 1

      long getCurrentRenderSteps();
         // Returns the current number of steps to be used to
         // completely render the shape.
         // Default: 1

      void setCurrentRenderSteps(long steps);
         // Sets the number of steps to be used to render
         // the shape. 1 <= steps <= max steps.

      virtual
      long render(long previousStep = 0);
         // Renders the shape. Returns the number of rendering
         // steps left until the object is completely rendered.
         // Returns 0 (zero) when finished rendering.
         // If previousStep is 0, the Shape starts rendering from the
         // beginning. If previousStep != 0, the Shape continues
         // rendering.
         // Psuedo code rendering loop...
         //    long lastStep = 0;
         //    while ((lastStep = shape.render(lastStep));

      virtual
      void renderLoRes();
         // Renders a "low res" or quick to render version of the shape.
         // Default behavior is to draw a wire frame bounding box.

      virtual
      void getBoundingBox(double& min_x, double& min_y, double& min_z,
                          double& max_x, double& max_y, double& max_z) const;
         // By default, the bounding box is (0,0,0),(0,0,0).

      virtual
      void getBoundingBox(sg_MinBoundBox& mbb) const;


   protected:
      double _min_x, _min_y, _min_z;
      double _max_x, _max_y, _max_z;
      long _maxRenderSteps;   // Default: 1
      long _curRenderSteps;   // Default: 1
     
};

class sg_Text : public sg_Shape
{
   // Draws text.

   public:

      sg_Text(long listbase,  // id first display list in font.
              const string& message) ;  // text to be displayed.

      long render(long previousStep = 0);

   private:

      long _lb;
      string _mess;
 
};

class sg_Cube : public sg_Shape
{
   // Draws a cube centered about the origin.

   public:

      sg_Cube(double edgeLength = 1.0);

      long render(long previousStep = 0);

   private:
 
      double _edgeLength;
};

class sg_Grid : public sg_Shape
{
   // Draws a grid, centered about the origin, in the XY plane.
   // 'width' is the distance along the X axis.
   // 'height' is the distance along the Y axis.
   // xsegs, ysegs are the number of subdivisions along the specified axis.
   // 1 <= xsegs,ysegs
   //
   // defaults:    width = 20000, height = 20000, xsegs = 20, ysegs = 20

   public:

      sg_Grid();

	  void setWidth(double width);
	  void setHeight(double height);
	  void setXsegs(double xsegs);
	  void setYsegs(double ysegs);
      long render(long previousStep = 0);
      void renderLoRes();

   private:
 
      double _width;
      double _height;
      double _xsegs;
      double _ysegs;
};

class sg_Cylinder : public sg_Shape
{
   // Draws a cylinder centered about the origin, pointing in +Z direction

   public:

      sg_Cylinder(double length = 1.0, double radius = 0.1, double faces = 10);
      ~sg_Cylinder();

      long render(long previousStep = 0);

   private:

      double _length, _radius, _faces;
      GLUquadricObj* _myQuad;
};

class sg_Cone : public sg_Shape
{
   // Draws a Cone centered about the origin, pointing in +Z direction

   public:

      sg_Cone(double length = 1.0, double radius = 0.1, double faces = 10);
      ~sg_Cone();

      long render(long previousStep = 0);

   private:

      double _length, _radius, _faces;
      GLUquadricObj* _myQuad;
};

class sg_Axis : public sg_Shape
{
   // Draws a 3d Coord Axis

   public:

      sg_Axis(float length, float radius, float faces, long lb);
      ~sg_Axis();

      long render(long previousStep = 0);

   private:

      GLUquadricObj* _myQuad;
      float _length, _radius, _faces;
      string _z, _y, _x;
      long _lb;
};



class sg_Box : public sg_Shape
{
   // Constructs an axis aligned box through two points. The points are assumed to be
   // opposite corners. Each axis is divided up into equal segments. The segments
   // control how many polygons will be generated for each face of the box.
   // For example: Specifiying (1,1,1) for the segments results in each face of the
   // cube having 1 polygon. Specifying (2,2,2) causes each face to have 4 polygons. 

   public:
      sg_Box(sg_Point const& a,
             sg_Point const& b,
             long x_segs = 1,
             long y_segs = 1,
             long z_segs = 1);

      long render(long previousStep = 0);

   private:

      void drawFace(double umin, double umax, double usegs,
                    double vmin, double vmax, double vsegs,
                    sg_Vector const& origin,
                    sg_Vector const& u_axis,
                    sg_Vector const& v_axis);

      sg_Point _a;
      sg_Point _b;
      long _x_segs;
      long _y_segs;
      long _z_segs;
};

class sg_PointLightShape : public sg_Shape
{
   // Wireframe representation of a point light

   public:
      sg_PointLightShape();
      long render(long previousStep = 0);
};

class sg_SpotLightShape : public sg_Shape
{
   // Wireframe representation of a spot light.

   public:
      sg_SpotLightShape();
      long render(long previousStep = 0);
};

class sg_DistantLightShape : public sg_Shape
{
   // Wireframe representation of a distant light.

   public:
      sg_DistantLightShape();
      long render(long previousStep = 0);

   private:

      void drawArrow();
};

class sg_ColorCube : public sg_Shape
{
   // Cube whose vertices are points in RGB color space.

   public:
      sg_ColorCube(double edgeLength = 1.0);
      long render(long previousStep = 0);

   private:
      double _edge;

};

class sg_WireSphere : public sg_Shape
{
   // Wire frame sphere. Polar axis aligned with z axis.

   public:
      sg_WireSphere(double radius,        // >= 0.0
                    long numMeridians,    // >= 0
                    long numSlices);      // >= 0
         // Radius is the radius of the sphere.
         // Meridians are "half great circles". They are half
         // circles drawn from pole to pole.
         // Slices are "latitude circles" and are centered
         // about the equator.

      long render(long previousStep = 0);

      void setRadius(double radius);
      void setMeridians(long numMeridians);
      void setMeridianSteps(long steps);
         // Controls how many line segments are used to
         // draw each meridian.
         // Default: 10.
      void setNumSlices(long numSlices);
      void setSliceSteps(long steps);
         // Controls how many line segments are used to
         // draw each slice.
         // Default: 20.

   private:
      double _radius;
      long _numMeridians;
      long _numMeridianSteps;
      long _numSlices;
      long _numSliceSteps;
};

class sg_AltAzIndicator : public sg_Shape
{
   public:
      sg_AltAzIndicator(double radius = 1.0);
      long render(long previousStep = 0);

      void setRadius(double radius);
      void setAltitude(double altitude);
      void setAzimuth(double azimuth);

      void setAzimColor(sg_RGBA const& color);
         // Default: sg_Color_Cyan

      void setAltColor(sg_RGBA const& color);
         // Default: sg_Color_Green

   private:
      double _radius;
      double _alt;
      double _azim;
      double _azimSteps; // Number of line segments use to draw each ring.
      double _altSteps;
      sg_RGBA _azimColor;
      sg_RGBA _altColor;

      void azimTickMarks(double angleIncr,
                         double tickMarkLen);
         // tickMarkLen is a percentage of _radius.

      void altTickMarks(double angleIncr,
                        double tickMarkLen);
};


class sg_LightBox : public sg_Shape
{
   public:
      sg_LightBox(double size = 1.0);
      long render(long previousStep = 1.0);

      void setBoxColor(sg_RGBA const& color);
      void setLightColor(sg_RGBA const& color);
         // This will be the emissive color of the light.

   private:
      double _size;
      sg_RGBA _boxColor;
      sg_RGBA _lightColor;
};

#endif  // INCLUDED_SG_SHAPE_HXX
