#include <stdlib.h>

#ifndef INCLUDED_SIO_CONVERTER_H
#include "io/sio_Converter.h"
#endif

#ifndef INCLUDED_SC_SUBFIELD_H
#include "container/sc_Subfield.h"
#endif


static const long TwoToTheEigth = 256;
static const long TwoToTheSixteenth = 65536;
static const long TwoToTheTwentyFourth = 16777216;

sio_Converter::~sio_Converter()
{
}


long
sio_Converter::makeVarSubfield(sc_Subfield* subfield,
                                   char const* data,
                                   long maxLength,
                                   char delimiter) const
{
   // This is the default implementation. Subclasses can override to make
   // it more efficient.

   long length = findVariableSubfieldLength(data,maxLength,delimiter);
   return makeFixedSubfield(subfield, data, length);
}

long
sio_Converter::findVariableSubfieldLength(char const* data,
                                   long maxLength,
                                   char delimiter) const
{
   // Length returned does not include the delimiter.

   char const* pos = data;
   long count = 0;

   while ((*pos != delimiter) && (count < maxLength))
      {
         pos++;
         count++;
      }

   return count;

}


// sio_Converter_A


long
sio_Converter_A::makeFixedSubfield(sc_Subfield* subfield,
                                       char const* data,
                                       long length) const
{
   // Convert the raw data into a string and create an sc_Subfield.

   char* tmpBuf = new char[length + 1];
   memcpy(tmpBuf,data,length);
   tmpBuf[length] = '\0';

   subfield->setA(tmpBuf);
   delete tmpBuf;

   return length;
}


// sio_Converter_I


long
sio_Converter_I::makeFixedSubfield(sc_Subfield* subfield,
                                       char const* data,
                                       long length) const
{
   // Convert the raw data into a long and create an sc_Subfield.

   char* tmpBuf = new char[length + 1];
   memcpy(tmpBuf,data,length);
   tmpBuf[length] = '\0';

   // XXX Should use strtol here for error checking.
   subfield->setI(atol(tmpBuf));
   delete tmpBuf;

   return length;
}


// sio_Converter_R


long
sio_Converter_R::makeFixedSubfield(sc_Subfield* subfield,
                                       char const* data,
                                       long length) const
{
   // Convert the raw data into a double and create an sc_Subfield.

   char* tmpBuf = new char[length + 1];
   memcpy(tmpBuf,data,length);
   tmpBuf[length] = '\0';

   // XXX Should use strtod here for error checking.
   subfield->setR(atof(tmpBuf));
   delete tmpBuf;

   return length;
}


// sio_Converter_S

long
sio_Converter_S::makeFixedSubfield(sc_Subfield* subfield,
                                       char const* data,
                                       long length) const
{
   // Convert the raw data into a double and create an sc_Subfield.

   char* tmpBuf = new char[length + 1];
   memcpy(tmpBuf,data,length);
   tmpBuf[length] = '\0';

   subfield->setS(atof(tmpBuf));
   delete tmpBuf;

   return length;
}

long
sio_Converter_BI8::makeFixedSubfield(sc_Subfield* subfield,
                                     char const* data,
                                     long length) const
{  // Convert a single character from the input data to an 8-bit signed integer
   if (length < 1)
      return 0;
  
   long val = static_cast<long>(data[0]);
   subfield->setBI8(val);

   return length;
}

long
sio_Converter_BUI8::makeFixedSubfield(sc_Subfield* subfield,
                                     char const* data,
                                     long length) const
{  // Convert a single character from the input data to an 8-bit unsigned integer
   if (length < 1)
      return 0;
  
   unsigned long val = static_cast<unsigned long>(data[0]);
   subfield->setBUI8(val);

   return length;
}

long
sio_Converter_BI16::makeFixedSubfield(sc_Subfield* subfield,
                                      char const* data,
                                      long length) const
{  // Convert two characters from the input data to a 16-bit signed integer
   if (length < 2)
      return 0;

   signed char MSB = static_cast<signed char>(data[0]);
   unsigned char LSB = static_cast<unsigned char>(data[1]);

   long val = static_cast<long>(MSB * TwoToTheEigth);
   val = val + LSB;
   subfield->setBI16(val);
	
   return length;
}

long
sio_Converter_BUI16::makeFixedSubfield(sc_Subfield* subfield,
                                      char const* data,
                                      long length) const
{  // Convert two characters from the input data to a 16-bit unsigned integer
   if (length < 2)
      return 0;

   unsigned char MSB = static_cast<unsigned char>(data[0]);
   unsigned char LSB = static_cast<unsigned char>(data[1]);

   unsigned long val = static_cast<unsigned long>(MSB * TwoToTheEigth);
   val = val + LSB;
   subfield->setBUI16(val);
	
   return length;
}

long
sio_Converter_BI24::makeFixedSubfield(sc_Subfield* subfield,
                                      char const* data,
                                      long length) const
{  // Convert three characters from the input data to a 24-bit signed integer
   if (length < 3)
      return 0;
   
   signed char MSB = static_cast<signed char>(data[0]);
   unsigned char LSB = static_cast<unsigned char>(data[2]);
   unsigned char OSB = static_cast<unsigned char>(data[1]);

   long val = static_cast<long>((MSB * TwoToTheSixteenth) + 
                                (OSB * TwoToTheEigth) +
                                LSB);
   subfield->setBI24(val);
   
   return length;
}                                                                            

long
sio_Converter_BUI24::makeFixedSubfield(sc_Subfield* subfield,
                                      char const* data,
                                      long length) const
{  // Convert three characters from the input data to a 24-bit unsigned integer
   if (length < 3)
      return 0;
   
   unsigned char MSB = static_cast<unsigned char>(data[0]);
   unsigned char LSB = static_cast<unsigned char>(data[2]);
   unsigned char OSB = static_cast<unsigned char>(data[1]);

   unsigned long val = static_cast<unsigned long>((MSB * TwoToTheSixteenth) + 
                                (OSB * TwoToTheEigth) +
                                LSB);
   subfield->setBUI24(val);
   
   return length;
}

long
sio_Converter_BI32::makeFixedSubfield(sc_Subfield* subfield,
                                      char const* data,
                                      long length) const
{  // Convert four characters from the input stream to a 32-bit signed integer
   if (length < 4)
      return 0;
	
   signed char MSB = static_cast<signed char>(data[0]);
   unsigned char OSB1 = static_cast<unsigned char>(data[1]);
   unsigned char OSB2 = static_cast<unsigned char>(data[2]);
   unsigned char LSB = static_cast<unsigned char>(data[3]);

   long val = static_cast<long>((MSB * TwoToTheTwentyFourth) +
                                (OSB1 * TwoToTheSixteenth) +
                                (OSB2 * TwoToTheEigth)  +
                                 LSB);

   subfield->setBI32(val);

   // the reader expects lengths in units of characters, not bits
   // XXX nuff said; this is a kludge
   return length ;
}

long
sio_Converter_BUI32::makeFixedSubfield(sc_Subfield* subfield,
                                      char const* data,
                                      long length) const
{  // Convert 4 characters from the input data to a 32-bit unsigned integer
   if (length < 4)
	  return 0;

   unsigned char MSB = static_cast<unsigned char>(data[0]);
   unsigned char OSB1 = static_cast<unsigned char>(data[1]);
   unsigned char OSB2 = static_cast<unsigned char>(data[2]);
   unsigned char LSB = static_cast<unsigned char>(data[3]);

   unsigned long val = static_cast<unsigned long>((MSB * TwoToTheTwentyFourth) +
                                (OSB1 * TwoToTheSixteenth) +
                                (OSB2 * TwoToTheEigth) + 
								 LSB);

   subfield->setBUI32(val);
   return length ;
}



