
#include <iostream>
#include <fstream>

#ifdef WIN32
   using namespace std;
#endif

#include "io/sio_8211DirEntry.h"
#include "io/sio_8211DDR.h"
#include "io/sio_8211DDRLeader.h"
#include "io/sio_8211DDRField.h"
#include "io/sio_8211DR.h"
#include "io/sio_8211DRLeader.h"
#include "io/sio_8211Utils.h"

ostream&
printDDRField(ostream& ostr, sio_8211DDRField const& field)
{
   ostr << "      DDR Field:\n";
   ostr << "         Data Struct Code: " << field.getDataStructCode() << '\n';
   ostr << "         Data Type Code: " << field.getDataTypeCode() << '\n';
   ostr << "         Field Name: '" << field.getDataFieldName() << "'\n";
   ostr << "         Array Descriptor: '" << field.getArrayDescriptor() << "'\n";
   ostr << "         Format Controls: '" << field.getFormatControls() << "'\n";
   ostr << '\n';

   return ostr;
}

ostream&
printField(ostream& ostr,
           sio_8211Field const& field)

{
   ostr << "      Length: " << field.getDataLength() << " '";

   char const* data = field.getData();
   for (long i = 0; i < field.getDataLength(); i++)
      {
         if ((data[i] >= 32) && (data[i] <= 126))
            ostr << static_cast<char>(data[i]);
         else
            if (data[i] == sio_8211UnitTerminator)
               ostr << '&';
            else if (data[i] == sio_8211FieldTerminator)
               ostr << ';';
            else
               ostr << '.';
      }
   ostr << '\n';
   return ostr;
}

ostream&
printDDRLeader(ostream& ostr, sio_8211DDRLeader const& leader)
{
   ostr << "DDR:\n\n";
   ostr << "   Leader:\n";
   ostr << "      Length: " << leader.getRecordLength() << "\n";
   ostr << "      Identifier: '" << leader.getLeaderIdentifier() << "'\n";
   ostr << "      Field Control Length: " << leader.getFieldControlLength() << '\n';
   ostr << "      Base Address of Field Area: " << leader.getBaseAddrOfFieldArea() << "\n";
   ostr << "      Size of Field Length Field: " << leader.getSizeOfFieldLengthField() << "\n";
   ostr << "      Size of Field Pos Field: " << leader.getSizeOfFieldPosField() << "\n";
   ostr << "      Size of Field Tag Field: " << leader.getSizeOfFieldTagField() << "\n";

   return ostr;
}

ostream&
printDRLeader(ostream& ostr, sio_8211DRLeader const& leader)
{
   ostr << "DR:\n\n";
   ostr << "   Leader:\n";
   ostr << "      Length: " << leader.getRecordLength() << "\n";
   ostr << "      Identifier: '" << leader.getLeaderIdentifier() << "'\n";
   ostr << "      Base Address of Field Area: " << leader.getBaseAddrOfFieldArea() << "\n";
   ostr << "      Size of Field Length Field: " << leader.getSizeOfFieldLengthField() << "\n";
   ostr << "      Size of Field Pos Field: " << leader.getSizeOfFieldPosField() << "\n";
   ostr << "      Size of Field Tag Field: " << leader.getSizeOfFieldTagField() << "\n";

   return ostr;
}


ostream&
printDirectory(ostream& ostr, sio_8211Directory const& dir)
{
   ostr << "\n   Directory Entries:\n\n";
   sio_8211Directory::const_iterator i;
   for (i = dir.begin(); i != dir.end(); i++)
      {
         ostr << "      Entry:\n";
         ostr << "         Field Tag: '" << (*i)->getTag() << "'\n";
         ostr << "         Field Length: " << (*i)->getFieldLength() << "\n";
         ostr << "         Field Position: " << (*i)->getPosition() << "\n";         
      }

   return ostr;
}

ostream&
printDR(ostream& ostr, sio_8211DR const& dr)
{
   sio_8211DRLeader const* leader = dynamic_cast<sio_8211DRLeader const*>(dr.getLeader());
   printDRLeader(ostr,*leader);

   sio_8211Directory const* dir = dr.getDirectory();
   printDirectory(ostr,*dir);

   sio_8211FieldArea const& fieldArea = dr.getFieldArea();
   ostr << "\n   Fields:\n\n";
   sio_8211FieldArea::const_iterator j;
   for (j = fieldArea.begin(); j != fieldArea.end(); j++)
      {
         printField(ostr,*j);
      }

   return ostr;
}

ostream&
printDDR(ostream& ostr, sio_8211DDR const& ddr)
{
   sio_8211DDRLeader const* leader = dynamic_cast<sio_8211DDRLeader const*>(ddr.getLeader());
   printDDRLeader(ostr,*leader);

   sio_8211Directory const* dir = ddr.getDirectory();
   printDirectory(ostr,*dir);

   sio_8211Directory::const_iterator j;
   for (j = dir->begin(); j != dir->end(); j++)
      {
         if (((*j)->getTag()).substr(0,3) != "000")
            {
               sio_8211DDRField ddf(*leader,*((*j)->getField()));
               printDDRField(ostr,ddf);
            }
         else
            printField(ostr,*((*j)->getField()));
      }

#if 0
   sio_8211FieldArea const& fieldArea = ddr.getFieldArea();
   ostr << "\n   Fields:\n\n";
   sio_8211FieldArea::const_iterator j;
   for (j = fieldArea.begin(); j != fieldArea.end(); j++)
      {
         printField(ostr,*j);
      }
#endif

   return ostr;
}



int
main(int argc, char** argv)
{

   if (argc != 2)
      {
         cerr << "Usage: " << argv[0] << " <8211 filename>\n";
         return 1;
      }

   ifstream istr(argv[1],ios::binary);
   if (!istr)
      {
         cerr << "Error opening file '" << argv[1] << "'. Aborting.\n";
         return 2;
      }


   cout << "\n\nFilename: '" << argv[1] << "'\n\n";

   sio_8211DDR ddr;
   istr >> ddr;
   printDDR(cout,ddr);

   sio_8211DRLeader const* leader;
   sio_8211DR dr;
   istr >> dr;
   while (!istr.eof())
      {
         leader = dynamic_cast<sio_8211DRLeader const*>(dr.getLeader());
         if (leader->getLeaderIdentifier() == 'R')
            dr.reuseLeaderAndDirectory(true);
         printDR(cout,dr);
         istr >> dr;
      }

   return 0;
}